import dune.codegen.blockstructured.quadrature
import dune.codegen.blockstructured.argument
import dune.codegen.blockstructured.geometry
import dune.codegen.blockstructured.spaces
import dune.codegen.blockstructured.basis
import dune.codegen.blockstructured.transformations
from dune.codegen.options import get_form_option
from dune.codegen.pdelab.quadrature import pymbolic_quadrature_position
from dune.codegen.blockstructured.spaces import lfs_inames
from dune.codegen.blockstructured.basis import (pymbolic_reference_gradient,
                                                pymbolic_basis)
from dune.codegen.blockstructured.geometry import (pymbolic_jacobian_inverse,
                                                   pymbolic_jacobian_determinant,
                                                   pymbolic_facet_jacobian_determinant,
                                                   to_global)
from dune.codegen.blockstructured.tools import sub_element_inames
from dune.codegen.pdelab import PDELabInterface


class BlockStructuredInterface(PDELabInterface):
    def __init__(self):
        PDELabInterface.__init__(self)

    def generate_accumulation_instruction(self, expr, visitor):
        if get_form_option('vectorization_blockstructured'):
            from dune.codegen.blockstructured.accumulation import generate_accumulation_instruction
            return generate_accumulation_instruction(expr, visitor)
        else:
            return PDELabInterface.generate_accumulation_instruction(self, expr, visitor)
    #
    # Local function space related generator functions
    #
    # TODO current way to squeeze subelem iname in, not really ideal

    def lfs_inames(self, element, restriction, number=None, context=''):
        return sub_element_inames() + lfs_inames(element, restriction, number, context)

    #
    # Test and trial function related generator functions
    #

    def pymbolic_basis(self, element, restriction, number, context=''):
        return pymbolic_basis(element, restriction, number, context)

    def pymbolic_reference_gradient(self, element, restriction, number, context=''):
        return pymbolic_reference_gradient(element, restriction, number, context)

    #
    # Geometry related generator functions
    #

    def pymbolic_spatial_coordinate(self):
        return to_global(pymbolic_quadrature_position())

    def pymbolic_jacobian_determinant(self):
        return pymbolic_jacobian_determinant()

    def pymbolic_jacobian_inverse(self, i, j, restriction):
        return pymbolic_jacobian_inverse(i, j, restriction)

    def pymbolic_facet_jacobian_determinant(self):
        return pymbolic_facet_jacobian_determinant()
