from dune.codegen.generation import (backend,
                                     kernel_cached,
                                     valuearg, instruction, globalarg)
from dune.codegen.options import get_form_option
from dune.codegen.pdelab.argument import CoefficientAccess
from dune.codegen.blockstructured.tools import micro_index_to_macro_index, sub_element_inames
from dune.codegen.pdelab.geometry import world_dimension
from loopy.types import NumpyType
import pymbolic.primitives as prim


def name_alias(container, lfs, element):
    name = container + "_" + lfs.name + "_alias"
    k = get_form_option("number_of_blocks")
    p = element.degree()
    dim = world_dimension()
    element_stride = tuple(p * (p * k + 1)**i for i in range(0, dim))
    index_stride = tuple((p * k + 1)**i for i in range(0, dim))
    globalarg(name, shape=(k,) * dim + (p + 1,) * dim, strides=element_stride + index_stride, managed=True)
    code = "const auto {} = &{}({},0);".format(name, container, lfs.name)
    instruction(within_inames=frozenset(),
                code=code,
                read_variables=frozenset({container}),
                assignees=frozenset({name}))
    return name


# TODO remove the need for element
@kernel_cached
@backend(interface="pymbolic_coefficient", name="blockstructured")
def pymbolic_coefficient(container, lfs, element, index):
    # TODO introduce a proper type for local function spaces!
    if isinstance(lfs, str):
        valuearg(lfs, dtype=NumpyType("str"))

    # If the LFS is not yet a pymbolic expression, make it one
    if not isinstance(lfs, prim.Expression):
        lfs = prim.Variable(lfs)

    # use higher order FEM index instead of Q1 index
    if get_form_option("vectorization_blockstructured"):
        subelem_inames = sub_element_inames()
        coeff_alias = name_alias(container, lfs, element)
        return prim.Subscript(prim.Variable(coeff_alias), tuple(prim.Variable(i) for i in subelem_inames + index))
    else:
        return prim.Call(CoefficientAccess(container), (lfs, micro_index_to_macro_index(element, index),))
